/**
 * \file main_pmsm.c
 * \author Martin Prudek
 * \brief Mainfile pro pmsm control.
 */

#ifndef NULL
#define NULL (void*) 0
#endif /*NULL*/


#include <stdlib.h> 	/*exit*/
#include <signal.h> 	/*signal handler Ctrl+C*/
#include <stdio.h> 	/*printf*/
#include <sched.h>	/*sheduler*/
#include <unistd.h>	/*usleep*/
#include <pthread.h>	/*threads*/

#include "rpin.h"	/*gpclk*/
#include "rp_spi.h"	/*spi*/
#include "misc.h" 	/*structure for priorities*/


#define PRUM_PROUD 	2061
#define PRUM_SOUC	6183

#define PRIOR_KERN	50
#define PRIOR_HIGH	49
#define PRIOR_LOW	20

#define THREAD_SHARED 	0
#define INIT_VALUE	0 	/*init value for semaphor*/

struct sigaction sighnd; /*struktura pro signal handler*/
struct rpi_in data;

uint8_t test;
uint16_t pwm1, pwm2, pwm3;


/**
 * \brief Initilizes GPCLK.
 */
int clk_init()
{
	initialise(); /*namapovani gpio*/
	initClock(PLLD_500_MHZ, 10, 0);
	gpioSetMode(4, FSEL_ALT0);
	return 0;
}
/*
 * \brief Terminates GPCLK.
 */

inline void clk_disable(){
	termClock(0);
}

/**
 * \brief Signal handler pro Ctrl+C
 */
void sighnd_fnc(){
	spi_disable();
	clk_disable();
        printf("\nprogram bezpecne ukoncen\n");
        exit(0);
}

void substractOffset(struct rpi_in* data, struct rpi_in* offset){
	data->pozice_raw=data->pozice;
	data->pozice-=offset->pozice;
	return;
}
/*
 * pocita procentualni odchylku od prumerneho proudu
 */
float diff_p(float value){
	return ((float)value-PRUM_PROUD)*100/PRUM_PROUD;
}
/*
 * pocita procentualni odchylku od prumerneho souctu proudu
 */
float diff_s(float value){
	return ((float)value-PRUM_SOUC)*100/PRUM_SOUC;
}
/*
 * tiskne potrebna data
 */
void printData(struct rpi_in data){
	float cur0, cur1, cur2;
	int i;
	if (data.adc_m_count){
		cur0=data.ch0/data.adc_m_count;
		cur1=data.ch1/data.adc_m_count;
		cur2=data.ch2/data.adc_m_count;
	}
	for (i = 0; i < 16; i++) {
			if (!(i % 6))
				puts("");
			printf("%.2X ", data.debug_rx[i]);
	}
	puts("");
	printf("\npozice=%d\n",(int32_t)data.pozice);
	printf("raw_pozice=%d\n",(int32_t)data.pozice_raw);
	printf("raw_pozice last11=%u\n",(data.pozice_raw&0x7FF));
	printf("index position=%d\n",(int16_t)data.index_position);
	printf("hal1=%d, hal2=%d, hal3=%d\n",data.hal1,data.hal2,data.hal3);
	printf("en1=%d, en2=%d, en3=%d (Last sent)\n",!!(0x40&test),!!(0x20&test),!!(0x10&test));
	printf("shdn1=%d, shdn2=%d, shdn3=%d (L.s.)\n",!!(0x08&test),!!(0x04&test),!!(0x02&test));
	printf("PWM1=%u(L.s.)\n",pwm1);
	printf("PWM2=%u(L.s.)\n",pwm2);
	printf("PWM3=%u(L.s.)\n",pwm3);
	printf("Pocet namerenych proudu=%u\n",data.adc_m_count);
	printf("(pwm1) (ch1)=%d (avg=%4.0f) (%2.2f%%)\n",data.ch1,cur1,diff_p(cur1));
	printf("(pwm2) (ch2)=%d (avg=%4.0f)(%2.2f%%)\n",data.ch2,cur2,diff_p(cur2));
	printf("(pwm3) (ch0)=%d (avg=%4.0f)(%2.2f%%)\n",data.ch0,cur0,diff_p(cur0));
	printf("soucet prumeru=%5.0f (%2.2f%%)\n",cur0+cur1+cur2,diff_s(cur0+cur1+cur2));
}
void prepare_tx(uint8_t * tx){

	/*Data format:
	 * tx[4] - bity 95 downto 88 - bits that are sent first
	 * tx[5] - bity 87 downto 80
	 * tx[6] - bity 79 downto 72
	 * tx[7] - bity 71 downto 64
	 * tx[8] - bity 63 downto 56
	 * tx[9] - bity 55 downto 48
	 * tx[10] - bity 47 downto 40
	 * tx[11] - bity 39 downto 32
	 * tx[12] - bity 31 downto 24
	 * tx[13] - bity 23 downto 16
	 * tx[14] - bity 15 downto 8
	 * tx[15] - bity 7 downto 0
	 *
	 * bit 95 - ADC reset
	 * bit 94 - enable PWM1
	 * bit 93 - enable PWM2
	 * bit 92 - enable PWM3
	 * bit 91 - shutdown1
	 * bit 90 - shutdown2
	 * bit 89 - shutdown3
	 * 	.
	 * 	.
	 * 	.
	 * bits 66 .. 56 - match PWM1
	 * bits 55 .. 45 - match PWM2
	 * bit 11,12 - Unused
	 * bits 42 .. 32  - match PWM3
	 */


	uint16_t tmp;

	/* keep the cap*/
	if (pwm1>2047) pwm1=2047;
	if (pwm2>2047) pwm2=2047;
	if (pwm3>2047) pwm3=2047;

	tx[0]=test; /*bit 94 - enable PWM1*/

	/*pwm1*/
	tx[7]=(tx[7] & 0xF8) | (0x07 & ((uint8_t*)&pwm1)[1]); /*MSB*/
	tx[8]=((uint8_t*)&pwm1)[0]; /*LSB*/

	/*pwm2*/
	tmp=pwm2;
	tmp<<=5;
	tx[9]=((uint8_t*)&tmp)[1]; /*MSB*/
	tx[10]=(tx[10] & 0x1F) | (0xE0 & ((uint8_t*)&tmp)[0]); /*LSB*/

	/*pwm3*/
	tx[10]=(tx[10] & 0xF8) | (0x07 & ((uint8_t*)&pwm3)[1]); /*MSB*/
	tx[11]=((uint8_t*)&pwm3)[0]; /*LSB*/


}
/**
 * Funkce pravidelne vypisuje posledni zjistenou pozici lokalniho motoru
 */
void * pos_monitor(void* param){
	set_priority(param);		/*set priority*/
	while(1){
		printData(data);
		usleep(1000000); 	/*1 Hz*/
	}
	return (void*)0;
}

/**
 * Funkce pravidelne vycita data z motoru
 */
void * read_data(void* param){
	int i;
	struct rpi_in pocatek;
	uint8_t tx[]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0} ;
	set_priority(param);				/*set priority*/
	pocatek = spi_read(tx);
		while(1){
			prepare_tx(tx);
			data = spi_read(tx);
			substractOffset(&data,&pocatek);
			usleep(1000);				/*1kHz*/
		}
}

/**
 * \brief Main function.
 */

int main(){
	uint16_t tmp;

	/*nastaveni priorit vlaken*/
	struct thread_param tsp;
	tsp.sch_policy = SCHED_FIFO;

	/*nastaveni signalu pro vypnuti pomoci Ctrl+C*/
	sighnd.sa_handler=&sighnd_fnc;
	sigaction(SIGINT, &sighnd, NULL );

	clk_init(); 		/* inicializace gpio hodin */
	spi_init(); 		/* iniicializace spi*/

	/*semafor pro detekci zpracovani parametru vlaken*/
	sem_init(&thd_par_sem,THREAD_SHARED,INIT_VALUE);

	/*vlakna*/
	pthread_t tid;			/*identifikator vlakna*/
	pthread_attr_t attr; 		/*atributy vlakna*/
	pthread_attr_init(&attr);	/*inicializuj implicitni atributy*/



	/*ziskavani dat z motoru*//*vysoka priorita*/
	tsp.sch_prior = PRIOR_HIGH;
	pthread_create(&tid, &attr, read_data, (void*)&tsp);

	/*vypisovani lokalni pozice*//*nizka priorita*/
	tsp.sch_prior = PRIOR_LOW;
	pthread_create(&tid, &attr, pos_monitor, (void*)&tsp);


	/*muzeme zavrit semafor*/
	sem_destroy(&thd_par_sem);

	while (1){
		scanf("%u",&tmp);
		printf("volba=%x\n",tmp);
		switch (tmp){
		case 1:
			scanf("%u",&pwm1);
			break;
		case 2:
			scanf("%u",&pwm2);
			break;
		case 3:
			scanf("%u",&pwm3);
			break;
		case 4:
			scanf("%u",&test);
			break;

		default:
			break;
		}

	}
	return 0;
}
